# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
import torch
import torch.nn as nn

import utils
from drqv2 import RandomShiftsAug, Encoder, Actor, Critic

class GPCritic(nn.Module):
    def __init__(self, num_critics, repr_dim, action_shape, feature_dim,
                 hidden_dim, ):
        super().__init__()

        self.trunk = nn.Sequential(nn.Linear(repr_dim, feature_dim),
                                   nn.LayerNorm(feature_dim), nn.Tanh())

        self.QS = nn.Sequential(
            utils.DenseParallel(feature_dim + action_shape[0], hidden_dim,
                                num_critics),
            nn.ReLU(inplace=True),
            utils.DenseParallel(hidden_dim, hidden_dim, num_critics),
            nn.ReLU(inplace=True),
            utils.DenseParallel(hidden_dim, 1, num_critics))

        self.apply(utils.weight_init)
        self.num_uncertainty_diffs = num_critics ** 2 - num_critics

    def forward(self, obs, action):
        h = self.trunk(obs)
        h_action = torch.cat([h, action], dim=-1)
        qs = self.QS(h_action)  # num_critics x bs x 1
        return torch.squeeze(torch.transpose(qs, 0, 1), dim=-1)  # bs x num_critics

    def get_pred_target_and_uncertainty(self, obs, action):
        qs = self.__call__(obs, action)
        q_pred = torch.mean(qs, dim=1, keepdim=True)  # bs x 1
        q_diffs = torch.abs(qs.unsqueeze(-1) - qs.unsqueeze(-2))  # bs x n_critic x n_critic
        q_unc = q_diffs.sum(-1).sum(-1, keepdim=True) / self.num_uncertainty_diffs  # bs x 1
        return q_pred, q_unc

    def compute_loss_part(self, obs, action, target_Q):
        QS = self.__call__(obs, action)  # bs x num_critics
        Q_errors = QS - target_Q
        loss = Q_errors.pow(2).sum(1).mean()
        return QS, loss, Q_errors

    def compute_loss(self, obs, action, reward, discount, next_obs,
                     next_action, target_critic, unc_coeff):
        with torch.no_grad():
            target_Q_mean, target_Q_unc = target_critic.get_pred_target_and_uncertainty(next_obs, next_action)
            target_V = target_Q_mean - unc_coeff * target_Q_unc
            target_Q = reward + (discount * target_V)
        QS = self.__call__(obs, action)  # bs x num_critics
        Q_errors = QS - target_Q
        loss = Q_errors.pow(2).sum(1).mean()
        return target_Q, QS, loss, Q_errors


class GPDrQV2Agent:
    def __init__(self, obs_shape, action_shape, device, lr, feature_dim,
                 hidden_dim, critic_target_tau, num_expl_steps,
                 update_every_steps, stddev_schedule, stddev_clip, use_tb,

                 unc_coeff, act_unc_coeff, tune_unc_coeff, tie_unc_coeff,
                 min_unc_coeff, max_unc_coeff, act_unc_coeff_shift_schedule,

                 unc_coeff_lr=1e-4,
                 unc_coeff_beta1=0.5,
                 unc_coeff_update_frequency=1
                 ):

        self.device = device
        self.critic_target_tau = critic_target_tau
        self.update_every_steps = update_every_steps
        self.use_tb = use_tb
        self.num_expl_steps = num_expl_steps
        self.stddev_schedule = stddev_schedule
        self.stddev_clip = stddev_clip

        self.instantiate_models(obs_shape, action_shape, device, feature_dim,
                                hidden_dim)
        # optimizers
        self.encoder_opt = torch.optim.Adam(self.encoder.parameters(), lr=lr)
        self.actor_opt = torch.optim.Adam(self.actor.parameters(), lr=lr)
        self.critic_opt = torch.optim.Adam(self.critic.parameters(), lr=lr)

        # data augmentation
        self.aug = RandomShiftsAug(pad=4)

        # uncertainty regularization
        self.unc_coeff = torch.tensor(unc_coeff).to(device)
        self.act_unc_coeff = torch.tensor(act_unc_coeff).to(device)
        self.tune_unc_coeff = tune_unc_coeff
        self.tie_unc_coeff = tie_unc_coeff
        self.unc_coeff_update_frequency = unc_coeff_update_frequency
        if tune_unc_coeff:
            self.unc_coeff.requires_grad = True
            self.unc_coeff_optimizer = torch.optim.Adam([self.unc_coeff],
                                                        lr=unc_coeff_lr,
                                                        betas=(unc_coeff_beta1, 0.999))
            if min_unc_coeff is not None and max_unc_coeff is not None:
                self.bound_unc_coeff = True
                self.min_unc_coeff = torch.tensor(min_unc_coeff).to(device)
                self.max_unc_coeff = torch.tensor(max_unc_coeff).to(device)
            else:
                self.bound_unc_coeff = False
            if self.tie_unc_coeff:
                self.act_unc_coeff_shift_schedule = act_unc_coeff_shift_schedule
        self.total_training_steps = 0

        self.train()
        self.critic_target.train()

    def instantiate_models(self, obs_shape, action_shape, device, feature_dim,
                           hidden_dim, ):
        # models
        self.encoder = Encoder(obs_shape).to(device)
        self.actor = Actor(self.encoder.repr_dim, action_shape, feature_dim,
                           hidden_dim).to(device)

        self.critic = Critic(self.encoder.repr_dim, action_shape, feature_dim,
                             hidden_dim).to(device)
        self.critic_target = Critic(self.encoder.repr_dim, action_shape,
                                    feature_dim, hidden_dim).to(device)
        self.critic_target.load_state_dict(self.critic.state_dict())

    def train(self, training=True):
        self.training = training
        self.encoder.train(training)
        self.actor.train(training)
        self.critic.train(training)

    def act(self, obs, step, eval_mode):
        obs = torch.as_tensor(obs, device=self.device)
        obs = self.encoder(obs.unsqueeze(0))
        stddev = utils.schedule(self.stddev_schedule, step)
        dist = self.actor(obs, stddev)
        if eval_mode:
            action = dist.mean
        else:
            action = dist.sample(clip=None)
            if step < self.num_expl_steps:
                action.uniform_(-1.0, 1.0)
        return action.cpu().numpy()[0]

    def get_qs_target_and_uncertainty(self, next_obs, next_action):
        target_Q1, target_Q2 = self.critic_target(next_obs, next_action)
        q_pred = (target_Q1 + target_Q2) / 2
        q_unc = torch.abs(target_Q1 - target_Q2)

        return q_pred, q_unc

    def get_qs_and_uncertainty(self, obs, action):
        Q1, Q2 = self.critic(obs, action)
        q_pred = (Q1 + Q2) / 2
        q_unc = torch.abs(Q1 - Q2)
        return q_pred, q_unc

    def update_unc_coeff(self, errors, step):
        metrics = dict()
        mean_error = errors.detach().mean()
        loss = mean_error * self.unc_coeff
        self.unc_coeff_optimizer.zero_grad(set_to_none=True)
        metrics['unccoeff_loss'] = loss
        metrics['unccoeff_value'] = self.unc_coeff
        metrics['unccoeff_error'] = mean_error
        loss.backward()
        self.unc_coeff_optimizer.step()
        if self.bound_unc_coeff:
            with torch.no_grad():
                self.unc_coeff.clamp_(min=self.min_unc_coeff, max=self.max_unc_coeff)
        if self.tie_unc_coeff:
            with torch.no_grad():
                if self.act_unc_coeff_shift_schedule is not None:
                    act_shift = utils.schedule(self.act_unc_coeff_shift_schedule, step)
                else:
                    act_shift = 0.0
                self.act_unc_coeff.data.copy_(self.unc_coeff.data + act_shift)
                metrics['unccoeff_actvalue'] = self.act_unc_coeff
        return metrics

    def update_critic(self, obs, action, reward, discount, next_obs, step,
                      tune_uncertainty_coefficient):
        metrics = dict()

        with torch.no_grad():
            stddev = utils.schedule(self.stddev_schedule, step)
            dist = self.actor(next_obs, stddev)
            next_action = dist.sample(clip=self.stddev_clip)
            target_Q_mean, target_Q_unc = self.get_qs_target_and_uncertainty(next_obs, next_action)
            target_V = target_Q_mean - self.unc_coeff * target_Q_unc
            target_Q = reward + (discount * target_V)

        Q1, Q2 = self.critic(obs, action)
        Q1_errors = Q1 - target_Q
        Q2_errors = Q2 - target_Q

        Q_errors = torch.cat([Q1_errors, Q2_errors], dim=1)

        critic_loss = Q_errors.pow(2).sum(1).mean()

        if self.use_tb:
            metrics['critic_target_q'] = target_Q.mean().item()
            metrics['critic_q1'] = Q1.mean().item()
            metrics['critic_q2'] = Q2.mean().item()
            metrics['critic_loss'] = critic_loss.item()

        # optimize encoder and critic
        self.encoder_opt.zero_grad(set_to_none=True)
        self.critic_opt.zero_grad(set_to_none=True)
        critic_loss.backward()
        self.critic_opt.step()
        self.encoder_opt.step()

        if tune_uncertainty_coefficient:
            metrics.update(self.update_unc_coeff(Q_errors, step))

        return metrics

    def update_actor(self, obs, step):
        metrics = dict()

        stddev = utils.schedule(self.stddev_schedule, step)
        dist = self.actor(obs, stddev)
        action = dist.sample(clip=self.stddev_clip)
        log_prob = dist.log_prob(action).sum(-1, keepdim=True)

        Q_mean, Q_unc = self.get_qs_and_uncertainty(obs, action)
        Q = Q_mean - self.act_unc_coeff * Q_unc

        actor_loss = -Q.mean()
        # optimize actor
        self.actor_opt.zero_grad(set_to_none=True)
        actor_loss.backward()
        self.actor_opt.step()

        if self.use_tb:
            metrics['actor_loss'] = actor_loss.item()
            metrics['actor_logprob'] = log_prob.mean().item()
            metrics['actor_ent'] = dist.entropy().sum(dim=-1).mean().item()

        return metrics

    def update(self, replay_iter, step):
        metrics = dict()

        if step % self.update_every_steps != 0:
            return metrics
        tune_uncertainty_coefficient = ((step % self.unc_coeff_update_frequency) == 0) and self.tune_unc_coeff

        batch = next(replay_iter)
        obs, action, reward, discount, next_obs = utils.to_torch(
            batch, self.device)

        # augment
        obs = self.aug(obs.float())
        next_obs = self.aug(next_obs.float())
        # encode
        obs = self.encoder(obs)
        with torch.no_grad():
            next_obs = self.encoder(next_obs)

        if self.use_tb:
            metrics['batch_reward'] = reward.mean().item()

        # update critic
        metrics.update(
            self.update_critic(obs, action, reward, discount, next_obs, step,
                               tune_uncertainty_coefficient=tune_uncertainty_coefficient))

        # update actor
        metrics.update(self.update_actor(obs.detach(), step))

        # update critic target
        utils.soft_update_params(self.critic, self.critic_target,
                                 self.critic_target_tau)

        return metrics


class ParallelGPDrQV2Agent(GPDrQV2Agent):
    def __init__(self, critic_type, num_critics, *args, **kwargs):
        self.critic_type = critic_type
        self.num_critics = num_critics
        self.num_uncertainty_diffs = self.num_critics ** 2 - self.num_critics
        super(ParallelGPDrQV2Agent, self).__init__(*args, **kwargs)

    def instantiate_models(self, obs_shape, action_shape, device, feature_dim,
                           hidden_dim, ):
        # models
        self.encoder = Encoder(obs_shape).to(device)
        self.actor = Actor(self.encoder.repr_dim, action_shape, feature_dim,
                           hidden_dim).to(device)
        if self.critic_type == 'dirac':
            self.critic = GPCritic(
                self.num_critics, self.encoder.repr_dim,
                action_shape, feature_dim, hidden_dim).to(device)
            self.critic_target = GPCritic(
                self.num_critics, self.encoder.repr_dim,
                action_shape, feature_dim, hidden_dim).to(device)
        else:
            raise NotImplementedError
        self.critic_target.load_state_dict(self.critic.state_dict())

    def get_qs_target_and_uncertainty(self, next_obs, next_action):
        return self.critic_target.get_pred_target_and_uncertainty(next_obs, next_action)

    def get_qs_and_uncertainty(self, obs, action):
        return self.critic.get_pred_target_and_uncertainty(obs, action)

    def update_critic(self, obs, action, reward, discount, next_obs, step,
                      tune_uncertainty_coefficient):
        metrics = dict()
        with torch.no_grad():
            stddev = utils.schedule(self.stddev_schedule, step)
            dist = self.actor(next_obs, stddev)
            next_action = dist.sample(clip=self.stddev_clip)

        target_Q, QS, critic_loss, Q_errors = self.critic.compute_loss(
            obs=obs, action=action, reward=reward, discount=discount,
            next_obs=next_obs, next_action=next_action,
            target_critic=self.critic_target, unc_coeff=self.unc_coeff)

        if self.use_tb:
            metrics['critic_target_q'] = target_Q.mean().item()
            metrics['critic_q'] = QS.mean().item()
            metrics['critic_loss'] = critic_loss.item()

        # optimize encoder and critic
        self.encoder_opt.zero_grad(set_to_none=True)
        self.critic_opt.zero_grad(set_to_none=True)
        critic_loss.backward()
        self.critic_opt.step()
        self.encoder_opt.step()

        if tune_uncertainty_coefficient:
            metrics.update(self.update_unc_coeff(Q_errors, step))

        return metrics
