import os.path as osp
import json
import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp
import matplotlib.pyplot as plt
import gym

from sac_models import StochasticActor, Critic, SAC
from generalized_sac_models import GPSAC
from efficient_layers import Critics, EfficientParallelCritics, EfficientQuantileParallelCritics
from samplers import Sampler
from buffers import FastReplayBuffer
import time
import logz
from collections import OrderedDict
from layers_utils import make_simple_dense_layers, make_modern_arch, ParallelOrthogonal

tfl = tf.keras.layers
tfm = tf.keras.models
tfo = tf.keras.optimizers
tfi = tf.keras.initializers
tfd = tfp.distributions
tfr = tf.keras.regularizers

MBPO_TARGET_ENT = {'Hopper-v2':-1, 'HalfCheetah-v2':-3, 'Walker2d-v2':-3, 'Ant-v2':-4, 'Humanoid-v2':-2}


def make_models(task_params, training_params, agent_params):
    exp_name = task_params.get('exp_name', None)
    dir_prefix = task_params.get('dir_prefix', '')
    exp_suffix = task_params.get('exp_suffix', '')
    task_name = task_params.get('task_name', 'walk')
    seed = task_params.get('seed', 1)
    episode_limit = task_params.get('episode_limit', 1000)

    epochs = training_params.get('epochs', 100)
    test_runs_per_epoch = training_params.get('test_runs_per_epoch', 10)
    steps_per_epoch = training_params.get('steps_per_epoch', 1000)
    initial_random_samples = training_params.get('initial_random_samples', 5000)
    start_training = training_params.get('start_training', 5000)
    training_catchup = training_params.get('training_catchup', False)

    buffer_size = training_params.get('buffer_size', 1000000)
    batch_size = training_params.get('batch_size', 128)
    updates_per_step = training_params.get('updates_per_step', 1)
    actor_delay = training_params.get('actor_delay', updates_per_step)
    target_delay = training_params.get('target_delay', 1)

    actor_lr = training_params.get('actor_lr', 1e-3)
    actor_beta1 = training_params.get('actor_beta1', 0.9)
    clip_actor_gradients = training_params.get('clip_actor_gradients', False)

    critic_lr = training_params.get('critic_lr', 1e-3)
    critic_beta1 = training_params.get('critic_beta1', 0.9)

    entropy_coeff_lr = training_params.get('entropy_coeff_lr', 1e-4)
    entropy_coeff_beta1 = training_params.get('entropy_coeff_beta1', 0.5)

    # Agent parameters
    algo = agent_params.get('algo', 'SAC')
    gamma = agent_params.get('gamma', 0.99)
    q_polyak = agent_params.get('q_polyak', 0.99)

    actor_layers_dim = agent_params.get('actor_layers_dim', 256)
    critic_layers_dim = agent_params.get('critic_layers_dim', 256)
    small_gain_final_layer = agent_params.get('small_gain_final_layer', False)

    use_modern_sn_critic = agent_params.get('use_modern_sn_critic', False)
    num_additional_critic_blocks = agent_params.get('num_additional_critic_blocks', 2)
    critic_modern_bottleneck_dim = agent_params.get('critic_modern_bottleneck_dim', 512)
    critic_modern_ln = agent_params.get('critic_modern_ln', True)
    initial_entropy_coefficient = agent_params.get('initial_entropy_coefficient', 0.1)
    tune_entropy_coefficient = agent_params.get('tune_entropy_coefficient', True)
    target_entropy = agent_params.get('target_entropy', None)
    target_entropy_scaling = agent_params.get('target_entropy_scaling', 1)
    correct_logprob_for_action_scale = agent_params.get('correct_logprob_for_action_scale', False)

    # GPSAC
    num_critics = agent_params.get('num_critics', 2)
    fast_parallel_critics = agent_params.get('fast_parallel_critics', False) # TODO
    critics_type = agent_params.get('critics_type', 'det')
    uncertainty_coeff = agent_params.get('uncertainty_coeff', 0.5)
    actor_uncertainty_coeff = agent_params.get('actor_uncertainty_coeff', uncertainty_coeff)
    tune_uncertainty_coefficient = agent_params.get('tune_uncertainty_coefficient', False)
    uncertainty_coeff_lr = agent_params.get('uncertainty_coeff_lr', 1e-4)
    uncertainty_coeff_beta1 = agent_params.get('uncertainty_coeff_beta1', 0.5)
    tie_actor_uncertainty_coeff = agent_params.get('tie_actor_uncertainty_coeff', False)
    actor_optimistic_shift_value = agent_params.get('actor_optimistic_shift_value', 0.0)
    actor_optimistic_shift_value_decay_steps = agent_params.get('actor_optimistic_shift_value_decay_steps', -1)
    tune_uncertainty_coefficient_delay = agent_params.get('tune_uncertainty_coefficient_delay', actor_delay)

    # QRSAC
    num_quantiles = agent_params.get('num_quantiles', 100)
    kappa = agent_params.get('kappa', 1.0)
    
    save_training_statistics = agent_params.get('save_training_statistics', False)


    base_gain = np.sqrt(2)
    orthogonal_init = tfi.Orthogonal(gain=base_gain)
    if fast_parallel_critics:
        cri_orthogonal_init = ParallelOrthogonal(gain=base_gain)
    else:
        cri_orthogonal_init = tfi.Orthogonal(gain=base_gain)

    if small_gain_final_layer:
        orthogonal_init_last = tfi.Orthogonal(gain=0.01 * base_gain)
        if fast_parallel_critics:
            cri_orthogonal_init_last = ParallelOrthogonal(gain=0.01*base_gain)
        else:
            cri_orthogonal_init_last = tfi.Orthogonal(gain=0.01 * base_gain)
    else:
        orthogonal_init_last = orthogonal_init
        cri_orthogonal_init_last = cri_orthogonal_init

    env = gym.make(task_name)
    eval_env = gym.make(task_name)

    action_size = tf.reduce_prod(env.action_space.shape)
    action_scale = env.action_space.high[0]

    if target_entropy == None:
        target_entropy = float(-1 * action_size)
        
    elif target_entropy == 'mbpo':
        target_entropy = MBPO_TARGET_ENT[task_name]

    target_entropy = target_entropy * target_entropy_scaling

    def make_actor():
        actor = StochasticActor([tfl.Dense(actor_layers_dim, 'relu', kernel_initializer=orthogonal_init),
                                 tfl.Dense(actor_layers_dim, 'relu', kernel_initializer=orthogonal_init),
                                 tf.keras.layers.Dense(action_size * 2, kernel_initializer=orthogonal_init_last)],
                                action_scale=action_scale,
                                correct_logprob_for_action_scale=correct_logprob_for_action_scale)
        return actor

    critic_out_dims = 1
    if fast_parallel_critics:
        parallel_models = num_critics
    else:
        parallel_models = None
    if critics_type == 'quant':
        critic_out_dims = num_quantiles
    if use_modern_sn_critic:
        def make_layers():
            return make_modern_arch(num_modern_blocks=num_additional_critic_blocks,
                                    dims=critic_layers_dim,
                                    bottleneck_dims=critic_modern_bottleneck_dim,
                                    init=cri_orthogonal_init,
                                    final_init=cri_orthogonal_init_last,
                                    out_dim=critic_out_dims,
                                    layer_norm=critic_modern_ln,
                                    parallel_models=parallel_models)
    else:
        def make_layers():
            return make_simple_dense_layers(num_hidden=1, dims=critic_layers_dim,
                                            nonlinearity='relu',
                                            init=cri_orthogonal_init,
                                            final_init=cri_orthogonal_init_last,
                                            out_dim=critic_out_dims,
                                            sn_hidden_layers=False,
                                            sn_first_layer=False,
                                            sn_last_layer=False,
                                            regularizer=None,
                                            parallel_models=parallel_models)
    if critics_type == 'quant':
        assert fast_parallel_critics, 'GP-QR-SAC is only implemented for the parallel ensemble implementation'
        def make_critic():
            critic = EfficientQuantileParallelCritics(make_layers())
            return critic
    elif fast_parallel_critics:
        def make_critic():
            critic = EfficientParallelCritics(make_layers())
            return critic
    else:
        def make_critic():
            critic = Critics(
                [make_layers() for _ in range(num_critics)])
            return critic


    sampler = Sampler(env, eval_env, episode_limit=episode_limit, init_random_samples=initial_random_samples)

    replay_buffer = FastReplayBuffer(buffer_size, visual_data=False)

    actor_optimizer = tf.keras.optimizers.Adam(actor_lr, beta_1=actor_beta1)
    critic_optimizer = tf.keras.optimizers.Adam(critic_lr, beta_1=critic_beta1)
    entropy_optimizer = tf.keras.optimizers.Adam(entropy_coeff_lr, beta_1=entropy_coeff_beta1)
    uncertainty_coeff_optimizer = tf.keras.optimizers.Adam(uncertainty_coeff_lr, beta_1=uncertainty_coeff_beta1)

    if algo == 'SAC':
        agent = SAC(make_actor=make_actor,
                    make_critics=make_critic,
                    make_critic2=make_critic,
                    actor_optimizer=actor_optimizer,
                    critic_optimizer=critic_optimizer,
                    entropy_optimizer=entropy_optimizer,
                    gamma=gamma,
                    q_polyak=q_polyak,
                    entropy_coefficient=initial_entropy_coefficient,
                    tune_entropy_coefficient=tune_entropy_coefficient,
                    target_entropy=target_entropy,
                    clip_actor_gradients=clip_actor_gradients,
                    save_training_statistics=save_training_statistics,
                    )
    elif algo == 'GPSAC':
        agent = GPSAC(make_actor=make_actor,
                      make_critics=make_critic,
                      actor_optimizer=actor_optimizer,
                      critic_optimizer=critic_optimizer,
                      entropy_optimizer=entropy_optimizer,
                      gamma=gamma,
                      q_polyak=q_polyak,
                      entropy_coefficient=initial_entropy_coefficient,
                      tune_entropy_coefficient=tune_entropy_coefficient,
                      target_entropy=target_entropy,
                      clip_actor_gradients=clip_actor_gradients,
                      
                      critics_type=critics_type,

                      uncertainty_coeff=uncertainty_coeff,
                      actor_uncertainty_coeff=actor_uncertainty_coeff,
                      tune_uncertainty_coefficient=tune_uncertainty_coefficient,
                      uncertainty_coeff_optimizer=uncertainty_coeff_optimizer,
                      tie_actor_uncertainty_coeff=tie_actor_uncertainty_coeff,
                      actor_optimistic_shift_value=actor_optimistic_shift_value,
                      actor_optimistic_shift_value_decay_steps=actor_optimistic_shift_value_decay_steps,
                      tune_uncertainty_coefficient_delay=tune_uncertainty_coefficient_delay,

                      num_quantiles=num_quantiles,
                      kappa=kappa,
                      save_training_statistics=save_training_statistics,
                      )
    else:
        raise NotImplementedError

    obs = np.expand_dims(env.reset().astype('float32'), axis=0)
    agent(obs)
    agent.summary()
    return env, eval_env, agent, replay_buffer, sampler


def run_exp(task_params, training_params, agent_params, agent, replay_buffer, sampler):
    exp_name = task_params.get('exp_name', None)
    dir_prefix = task_params.get('dir_prefix', '')
    exp_suffix = task_params.get('exp_suffix', '')
    task_name = task_params.get('task_name', 'walk')
    seed = task_params.get('seed', 1)
    episode_limit = task_params.get('episode_limit', 1000)

    epochs = training_params.get('epochs', 100)
    test_runs_per_epoch = training_params.get('test_runs_per_epoch', 10)
    steps_per_epoch = training_params.get('steps_per_epoch', 1000)
    initial_random_samples = training_params.get('initial_random_samples', 5000)
    start_training = training_params.get('start_training', 5000)
    training_catchup = training_params.get('training_catchup', False)

    buffer_size = training_params.get('buffer_size', 1000000)
    batch_size = training_params.get('batch_size', 128)
    updates_per_step = training_params.get('updates_per_step', 1)
    actor_delay = training_params.get('actor_delay', updates_per_step)
    target_delay = training_params.get('target_delay', 1)

    actor_lr = training_params.get('actor_lr', 1e-3)
    actor_beta1 = training_params.get('actor_beta1', 0.9)
    clip_actor_gradients = training_params.get('clip_actor_gradients', False)

    critic_lr = training_params.get('critic_lr', 1e-3)
    critic_beta1 = training_params.get('critic_beta1', 0.9)

    entropy_coeff_lr = training_params.get('entropy_coeff_lr', 1e-4)
    entropy_coeff_beta1 = training_params.get('entropy_coeff_beta1', 0.5)

    # Agent parameters
    algo = agent_params.get('algo', 'SAC')
    gamma = agent_params.get('gamma', 0.99)
    q_polyak = agent_params.get('q_polyak', 0.99)

    actor_layers_dim = agent_params.get('actor_layers_dim', 256)
    critic_layers_dim = agent_params.get('critic_layers_dim', 256)
    small_gain_final_layer = agent_params.get('small_gain_final_layer', False)

    use_modern_sn_critic = agent_params.get('use_modern_sn_critic', False)
    num_additional_critic_blocks = agent_params.get('num_additional_critic_blocks', 2)
    critic_modern_bottleneck_dim = agent_params.get('critic_modern_bottleneck_dim', 512)
    critic_modern_ln = agent_params.get('critic_modern_ln', True)
    initial_entropy_coefficient = agent_params.get('initial_entropy_coefficient', 0.1)
    tune_entropy_coefficient = agent_params.get('tune_entropy_coefficient', True)
    target_entropy = agent_params.get('target_entropy', None)
    target_entropy_scaling = agent_params.get('target_entropy_scaling', 1)
    correct_logprob_for_action_scale = agent_params.get('correct_logprob_for_action_scale', False)

    # GPSAC
    num_critics = agent_params.get('num_critics', 2)
    fast_parallel_critics = agent_params.get('fast_parallel_critics', False)  # TODO
    critics_type = agent_params.get('critics_type', 'det')
    uncertainty_coeff = agent_params.get('uncertainty_coeff', 0.5)
    actor_uncertainty_coeff = agent_params.get('actor_uncertainty_coeff', uncertainty_coeff)
    tune_uncertainty_coefficient = agent_params.get('tune_uncertainty_coefficient', False)
    uncertainty_coeff_lr = agent_params.get('uncertainty_coeff_lr', 1e-4)
    uncertainty_coeff_beta1 = agent_params.get('uncertainty_coeff_beta1', 0.5)
    tie_actor_uncertainty_coeff = agent_params.get('tie_actor_uncertainty_coeff', False)
    actor_optimistic_shift_value = agent_params.get('actor_optimistic_shift_value', 0.0)
    actor_optimistic_shift_value_decay_steps = agent_params.get('actor_optimistic_shift_value_decay_steps', -1)
    tune_uncertainty_coefficient_delay = agent_params.get('tune_uncertainty_coefficient_delay', actor_delay)

    # QRSAC
    num_quantiles = agent_params.get('num_quantiles', 100)
    kappa = agent_params.get('kappa', 1.0)

    save_training_statistics = agent_params.get('save_training_statistics', False)

    description_prefix = 'proprio'
    brief_exp_description = '{}_{}_{}b'.format(algo, description_prefix, batch_size)

    if algo == 'GPSAC':
        brief_exp_description += '{}_{}cri_{}beta'.format(critics_type, num_critics, uncertainty_coeff)

    if updates_per_step > 1:
        brief_exp_description += '_x{}up'.format(updates_per_step)
        
    log_exp_name = exp_name or brief_exp_description
    log_exp_name += exp_suffix

    log_dir = osp.join('experiments_data/', '{}{}/{}/{}'.format(
        dir_prefix, task_name, log_exp_name, time.time()))

    params = OrderedDict({'task_params': task_params, 'training_params': training_params,
                          'agent_params': agent_params})

    logz.configure_output_dir(log_dir)
    logz.save_params(params)

    mean_test_returns = []
    mean_test_std = []
    steps = []

    step_counter = 0
    logz.log_tabular('epoch', 0)
    logz.log_tabular('number_collected_observations', step_counter)
    print('Epoch {}/{} - total steps {}'.format(0, epochs, step_counter))
    start_training_time = time.time()
    evaluation_stats = sampler.evaluate(agent, test_runs_per_epoch, log=False)
    for k, v in evaluation_stats.items():
        logz.log_tabular(k, v)
    if save_training_statistics:
        agent.save_metrics_and_reset()
        for k, v in agent._latest_log_dict.items():
            logz.log_tabular(k, v)
    mean_test_returns.append(evaluation_stats['episode_returns_mean'])
    mean_test_std.append(evaluation_stats['episode_returns_std'])
    steps.append(step_counter)
    epoch_end_eval_time = time.time()
    evaluation_time = epoch_end_eval_time - start_training_time

    logz.log_tabular('training_time', 0.0)
    logz.log_tabular('evaluation_time', np.around(evaluation_time, decimals=3))
    logz.dump_tabular()
    for e in range(epochs):
        logz.log_tabular('epoch', e + 1)
        epoch_start_time = time.time()
        while step_counter < (e + 1) * steps_per_epoch:
            traj_data = sampler.sample_steps(agent, 0.1, n_steps=1)
            replay_buffer.add(traj_data)
            step_counter += traj_data['n']
            if step_counter > start_training:
                agent.train(replay_buffer, batch_size=batch_size,
                            n_updates=updates_per_step * traj_data['n'],
                            act_delay=actor_delay,
                            tar_delay=target_delay)
            elif step_counter == start_training:
                if training_catchup:
                    n_updates = updates_per_step * start_training
                else:
                    n_updates = updates_per_step * traj_data['n']
                agent.train(replay_buffer, batch_size=batch_size,
                            n_updates=n_updates,
                            act_delay=actor_delay,
                            tar_delay=target_delay)

        logz.log_tabular('number_collected_observations', step_counter)
        epoch_end_training_time = time.time()
        training_time = epoch_end_training_time - epoch_start_time
        print('Epoch {}/{} - total steps {}'.format(e + 1, epochs, step_counter))
        evaluation_stats = sampler.evaluate(agent, test_runs_per_epoch, log=False)
        for k, v in evaluation_stats.items():
            logz.log_tabular(k, v)
        if save_training_statistics:
            agent.save_metrics_and_reset()
            for k, v in agent._latest_log_dict.items():
                logz.log_tabular(k, v)
        mean_test_returns.append(evaluation_stats['episode_returns_mean'])
        mean_test_std.append(evaluation_stats['episode_returns_std'])
        steps.append(step_counter)
        epoch_end_eval_time = time.time()
        evaluation_time = epoch_end_eval_time - epoch_end_training_time
        logz.log_tabular('training_time', np.around(training_time, decimals=3))
        logz.log_tabular('evaluation_time', np.around(evaluation_time, decimals=3))
        logz.dump_tabular()
    total_training_time = time.time() - start_training_time
    print('Total training time: {}'.format(total_training_time))
    plt.errorbar(steps, mean_test_returns, mean_test_std)
    plt.xlabel('steps')
    plt.ylabel('returns')
    plt.show(block=False)
    return mean_test_returns


def run_experiment_from_params(task_name=None, base_param_path='params.json', repetitions=1, gpu=0, memory_limit=10000):
    if gpu >= 0:
        GPU_TO_USE = gpu
        gpus = tf.config.experimental.list_physical_devices('GPU')
        if gpus:
            try:
                tf.config.experimental.set_visible_devices(gpus[GPU_TO_USE], 'GPU')
                tf.config.experimental.set_virtual_device_configuration(
                    gpus[GPU_TO_USE],
                    [tf.config.experimental.VirtualDeviceConfiguration(memory_limit=memory_limit)])
                logical_gpus = tf.config.experimental.list_logical_devices('GPU')
                print(len(gpus), "Physical GPUs,", len(logical_gpus), "Logical GPU")
            except RuntimeError as e:
                print(e)
    else:
        tf.config.set_visible_devices([], 'GPU')

    with open(base_param_path, 'r') as inp:
        data = json.loads(inp.read())
    task_params = data['task_params']
    training_params = data['training_params']
    agent_params = data['agent_params']

    if task_name is not None:
        task_params['task_name'] = task_name

    for i in range(repetitions):
        task_params['seed'] = np.random.randint(10000)
        env, eval_env, agent, replay_buffer, sampler = make_models(task_params, training_params, agent_params)
        run_exp(task_params, training_params, agent_params, agent, replay_buffer, sampler)
    return agent, replay_buffer, sampler
